<?php
// $Id: database.oracle.inc,v 1.1.2.23 2010/04/04 11:35:22 andreagariboldi Exp $

openlog("database.oracle.log", LOG_PID | LOG_PERROR, LOG_USER);

define('ORACLE_EMPTY_STRING_REPLACER','^'); // used to replace '' character in queries
define('ORACLE_IDENTIFIER_MAX_LENGTH',30); // maximum oracle identifier length (e.g. table names cannot exceed this length)  
define('ORACLE_LONG_IDENTIFIER_PREFIX','L#'); // prefix used for long identifier keys
define('ORACLE_BLOB_PREFIX','B^#'); // prefix used for BLOB values
define('ORACLE_MAX_VARCHAR2_LENGTH',4000); // maximum length for a string value in a table column in oracle 
define('ORACLE_ROWNUM_ALIAS','RWN_TO_REMOVE'); // alias used for queryRange filtering (we have to remove that from resultsets)
define('ORACLE_LONG_IDENTIFIER_PREFIX_LENGTH',strlen(ORACLE_LONG_IDENTIFIER_PREFIX));
define('ORACLE_BLOB_PREFIX_LENGTH',strlen(ORACLE_BLOB_PREFIX));
define('ORACLE_ROWNUM_ALIAS_LOWER',strtolower(ORACLE_ROWNUM_ALIAS));

$query_cache= array();

/**
 * @file
 * Database interface code for Oracle database servers.
 */

/**
 * @ingroup database
 * @{db
 */

function oracle_query($query,$args=NULL)
{
  global $active_db;

  if (!$active_db) syslog(LOG_ERR,$query);
  
  $stmt= $active_db->prepare($query);
  $stmt->execute($args);
  
  return $stmt;
}


/**
 * Report database status.
 */
function db_status_report() {
  $t = get_t();

  $version = db_version();

  $form['oracle'] = array(
    'title' => $t('Oracle database'),
    'value' => $version,
  );

  return $form;
}

/**
 * Returns the version of the database server currently in use.
 *
 * @return Database server version
 */
function db_version() {
  return db_result(db_query("select banner from v\$version where rownum < 2"));
}

/**
 * Initialize a database connection.
 */
function db_connect($url)
{  
  global $active_db;
  
  $url = parse_url($url);
  $conn_string = '';

  $url['user'] = urldecode($url['user']);
  $url['pass'] = isset($url['pass']) ? urldecode($url['pass']) : '';
  $url['host'] = urldecode($url['host']);
  $url['path'] = urldecode($url['path']);

  if (!isset($url['port'])) {
    $url['port'] = "1521";
  }

  if (!isset($url['host'])) {
    $url['host'] = "localhost";
  }

  $url['host'] = $url['host'] .':'. $url['port'];
  $url['host'] = "//" . $url['host'] . $url['path']; 

  $dsn = 'oci:dbname=' . $url['host'] . ';charset=AL32UTF8';
  
  //syslog(LOG_ERR,"dsn: ".$dsn);
  
  $conn= false;
  
  try
  {
	  $conn= new PDO($dsn, $url['user'], $url['pass']);

	  $conn->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);
	  $conn->setAttribute(PDO::ATTR_CASE, PDO::CASE_LOWER);
	  $conn->setAttribute(PDO::ATTR_STRINGIFY_FETCHES, TRUE);
	  
      $active_db= $conn;
	  
      oracle_query("ALTER SESSION SET cursor_sharing='FORCE'");
	  oracle_query("ALTER SESSION SET NLS_NUMERIC_CHARACTERS = '.,'");
	  
	  oracle_reset_long_identifiers(true);
  }
  catch (Exception $ex)
  {
      _db_error_page(htmlentities($ex->getMessage()));
  	  return false;
  } 
   
  return $conn;
}

function oracle_concat_bind_variable($match) {
  
  switch ($match[0]) {
    case '%d': 
      return "'||trim(to_char(%d))||'";
    case '%s':
      return "'||%s||'";
    case '%n':
      return "'||trim(to_char(%n))||'";
    case '%f':
      return "'||trim(to_char(%f))||'";
  }
  
}

function oracle_escape_reserved_words($query)
{
        $search = array ("/({)(\w+)(})/e", // escapes all table names
                         "/({L#)([0-9]+)(})/e", // escapes long id
		                 "/([\:])(uid|session|file|size|access|mode|comment)/e",
			   	         "/(<uid>|<session>|<file>|<size>|<access>|<mode>|<comment>)/e",
			   	         "/([\(\.\s,])(uid|session|file|size|access|mode|comment)([,\s\=)])/e",
                         "/([\(\.\s,])(uid|session|file|size|access|mode|comment)$/e");

	    $replace = array ("'\"\\1'.strtoupper('\\2').'\\3\"'",
				          "'\"\\1'.strtoupper('\\2').'\\3\"'",
				          "'\\1'.'db_'.'\\2'.'\\3'",
					      "strtoupper('\"\\1\"')",
					      "'\\1'.strtoupper('\"\\2\"').'\\3'",
	                      "'\\1'.strtoupper('\"\\2\"')");
	
    $query= preg_replace($search, $replace, $query);

    return $query;
}

function oracle_concat_bind_variables($match)
{
  return preg_replace_callback("/(%d|%s|%f|%n)/", 'oracle_concat_bind_variable', $match[0]);
}

function oracle_check_db_prefix($db_prefix)
{
    global $oracle_prefix;
    
    if (empty($db_prefix)) return;

    if (!is_array($oracle_prefix)) $oracle_prefix= array();

    if (!isset($oracle_prefix[$db_prefix]))
    {
    	oracle_query("begin identifier.check_db_prefix(?); end;",array($db_prefix));
    	$oracle_prefix[$db_prefix]= 'x';
    }
}

function oracle_prefix_tables($sql)
{
      global $db_prefix;
      
      if (empty($db_prefix))
	    return strtr($sql, array('{' => '', '}' => ''));
	  elseif (is_array($db_prefix))
	  {
		    if (array_key_exists('default', $db_prefix))
		    {
		      $tmp = $db_prefix;
		      unset($tmp['default']);
		      
		      foreach ($tmp as $key => $val)
		      {
		      	 oracle_check_db_prefix($val);
		         $sql = strtr($sql, array('{'. strtoupper($key) .'}' => (empty($val) ? strtoupper($key) : strtoupper($val) .'"."'. strtoupper($key))));
		      }
		      
		      oracle_check_db_prefix($db_prefix['default']);
		      return strtr($sql, array('{' => (empty($db_prefix['default']) ? '' : strtoupper($db_prefix['default']). '"."'), '}' => ''));
		    }
		    else
		    {
		      foreach ($db_prefix as $key => $val)
		      { 
		      	 oracle_check_db_prefix($val);
		         $sql = strtr($sql, array('{'. strtoupper($key) .'}' => (empty($val) ? strtoupper($key) : strtoupper($val) .'"."'. strtoupper($key))));
		      }
		      return strtr($sql, array('{' => '', '}' => ''));
		    }
	  }
	  else
	  {
	  	 oracle_check_db_prefix($db_prefix); 
	     return strtr($sql, array('{' =>  strtoupper($db_prefix) . '"."', '}' => ''));
	  }
	  
}

function oracle_escape_ansi($query)
{
     if (preg_match("/^select /i",$query)&&!preg_match("/^select(.*)from/ims",$query))
  	  $query.= ' FROM DUAL';

     $search = array ('/("\w+?")/e');

     $replace = array ("strtoupper('\\1')");

     return str_replace('\\"','"',preg_replace($search, $replace, $query));
}

function oracle_remove_long_identifiers($queryPart)
{
	global $searchLongIdentifiers,$replaceLongIdentifiers;
	
	if (count($searchLongIdentifiers))
	  return preg_replace($searchLongIdentifiers, $replaceLongIdentifiers, $queryPart);
	else
	  return $queryPart;
}

function oracle_escape_long_identifiers($query)
{
  	    $ret= "";
  	    
  	    // do not replace things in literals
  	    $literals= array(); 
  	    preg_match_all("/'.*?'/", $query,$literals);
  	    
  	    $literals= $literals[0];
  	    
        $replaceable = preg_split("/'.*?'/", $query);
        
        $lidx= 0;

        // assume that a query cannot start with a literal and that 
        foreach ($replaceable as $toescape)
             $ret.= oracle_remove_long_identifiers($toescape).(isset($literals[$lidx]) ? $literals[$lidx++] : "");          
        
        return $ret;  	
}

function oracle_escape_empty_literal($match)
{
	if ($match[0]=="''") 
	  return "'".ORACLE_EMPTY_STRING_REPLACER."'";
	else
	  return $match[0];
}

function oracle_escape_empty_literals($query)
{
  	    return preg_replace_callback("/'.*?'/", 'oracle_escape_empty_literal', $query);
}


function oracle_escape_group_by_alias_callback($match, $al= FALSE, $cl= FALSE, $init = FALSE)
{
  static $aliases;
  static $cols;
  
  if ($init) {
    $aliases= $al;
    $cols= $cl;
    return;
  }

  $retval= $match[0];
  $pieces= explode("ORDER BY",$retval); 
  $retval= $pieces[0];
  
  foreach ($aliases as $key => $alias)
    $retval= str_replace($alias, trim($cols[$key]), $retval);

  return $retval.( isset($pieces[1]) ? "ORDER BY".$pieces[1] : "");
}

function oracle_escape_group_by_alias($query)
{
	if (preg_match('/GROUP BY/',$query))
	{
        preg_match_all('/(SELECT COUNT\(\*\) FROM \()*(SELECT )*((.+?)\sAS\s([^,\s]+)[,\s]*?)/',$query,$matches);
        $cols= $matches[4];
        $aliases= $matches[5];
        oracle_escape_group_by_alias_callback(NULL,$aliases,$cols,TRUE); 
        $query= preg_replace_callback("/GROUP BY.*/",'oracle_escape_group_by_alias_callback',$query);
	}
	
	return $query;
}

function oracle_escape_if_function($query)
{
	return preg_replace("/IF\s*\((.*?),(.*?),(.*?)\)/", 'case when \1 then \2 else \3 end', $query);
}

function oracle_escape_multiple_insert($query)
{
	if (preg_match('/^INSERT.*\)\s*,\s*\(/si',$query))
	{
		$matches= array();
		preg_match('/^INSERT.*VALUES/si',$query,$matches);
		$insert_prefix= $matches[0];
		preg_match_all('/\(.*?\)/si',$query,$matches);
		array_shift($matches[0]);
		
		$query= "begin ";
		
		foreach ($matches[0] as $match)
		  $query.= $insert_prefix." ".$match."; ";
		
		$query.= "end;";
	}
	
	return $query;
}

function oracle_escape_truncate($query)
{
	return str_replace('TRUNCATE "','TRUNCATE TABLE "', $query); 
}

function oracle_format_query($query)
{
    global $oracle_unformatted_query, $query_cache, $oracle_use_memcached;
    
    $oracle_unformatted_query= $query;
    $iquery= md5(oracle_prefix_tables($query));
   
    if (isset($query_cache[$iquery]))
      return $query_cache[$iquery];
    
    if ($oracle_use_memcached)
    {
       $cached= cache_get($iquery, 'cache_oracle'); 
       if ($cached)
       {
          $query_cache[$iquery]= $cached->data;
          return $cached->data;
       }
    }
    
	/* escape ansi sql */
    $query= oracle_escape_ansi($query);
	
	/* escape long identifiers */
    $query= oracle_escape_long_identifiers($query);
    
    /* escape reserved words */
    $query= oracle_escape_reserved_words($query);

    /* escape empty literals */
    $query= oracle_escape_empty_literals($query);
    
    /* escape multiple insert (...) values (...),(...),(...) */
    $query= oracle_escape_multiple_insert($query);
    
    /* escape bind variables contained into quotes like '%s' */
	$query= preg_replace_callback("/'.*?'/",'oracle_concat_bind_variables',$query);
	
	/* escape IF function to case block */
	$query= oracle_escape_if_function($query);
	
	/* escape GROUP BY using aliases instead of selected columns */
	$query= oracle_escape_group_by_alias($query);

	/* escape TRUNCATE */
	$query= oracle_escape_truncate($query);
	
	/* other generic rules */
	$search = array ("''||", // remove empty concatenations leaved by concatenate_bind_variables
                     "||''",
                     "IN ()", // translate 'IN ()' to '= NULL' they do not match anything anyway (always false)
                     '(FALSE)',
                     'POW(',
                     ") AS count_alias", // hacks
                     'ON (nid) * FROM "{NODE}"',
                     'AS _random',
                     'ORDER BY _random');

	$replace = array ("",
                      "",
                      "= NULL",
                      "(1=0)",
                      "POWER(",
                      ") count_alias",
                      'x.* FROM "{NODE}" x',
                      'AS "_random"',
                      'ORDER BY "_random"');

	$query= str_replace($search, $replace, $query);
	$query= oracle_prefix_tables($query);

    if ($oracle_use_memcached)
      cache_set($iquery,$query,'cache_oracle');
      
    $query_cache[$iquery]= $query;
    
    return $query;
}

function oracle_db_query_callback($match, $init = FALSE, $count=FALSE) {
  static $args = NULL;
  static $cnt = -1;
  
  if ($count) return $cnt;
  
  if ($init) {
    $args = $match;
    $cnt = -1;
    return;
  }
  
  $cnt = $cnt + 1;

  switch ($match[1]) {
    case '%d': // We must use type casting to int to convert FALSE/NULL/(TRUE?)
      return ':d'. $cnt;
    case '%s':
      return ':s'. $cnt;
    case '%n':
      return ':n'. $cnt;
    case '%%':
      $cnt = $cnt - 1;
      return '%';
    case '%f':
      return ':f'. $cnt;
    case '%b': // binary data
      return ':b'. $cnt;
  }
}

function db_escape_string($string)
{
   return str_replace("'", "''", $string);
}

function oracle_string_to_stream($value)
{
     $stream = fopen('php://memory', 'a');
     fwrite($stream, $value);
     rewind($stream);
  	 return $stream;
}

function oracle_write_blob($value)
{
	   global $active_db;
	   
  	   $hash= md5($value);
  	   $handle= 0;
  	   
  	   $stream= oracle_string_to_stream($value);
  	   
  	   $stmt= $active_db->prepare("begin identifier.write_blob(?,?,?); end;");
  	   
  	   $stmt->bindParam(1, $hash, PDO::PARAM_STR, 32);
       $stmt->bindParam(2, $handle, PDO::PARAM_INT|PDO::PARAM_INPUT_OUTPUT, 32);
       $stmt->bindParam(3, $stream, PDO::PARAM_LOB);

       $stmt->execute();
       
       $handle= ORACLE_BLOB_PREFIX.$handle;
       
       return $handle;
}
  
function oracle_read_blob($handle)
{
	  global $active_db;
	  
	  $handle= (int)substr($handle,strlen(ORACLE_BLOB_PREFIX));
  	  $stmt= $active_db->prepare("select content from blobs where blobid= ?");
      $stmt->bindParam(1, $handle, PDO::PARAM_INT, 32);
      $stmt->execute();
 	  return $stmt->fetchColumn();
}

function oracle_cleanup_arg_value($value)
{
  	    if (is_string($value))
  	    {
  	      if ($value=='')
  	        return ORACLE_EMPTY_STRING_REPLACER;
  	      elseif (strlen($value)>ORACLE_MAX_VARCHAR2_LENGTH)
  	        return oracle_write_blob($value);
  	      else
  	        return $value;
  	    }
  	    else
  	      return $value;
}

function oracle_bind_query_callback($match, $pstmt= NULL, $init = FALSE, $qry = "") {
  static $args = NULL;
  static $stmt = NULL;
  static $cnt = -1;
  static $query = "";


  if ($init) {
    $args = $match;
    $stmt = $pstmt;
    $cnt = -1;
    $query = $qry;
    return;
  }
  
  $cnt = $cnt + 1;


  switch ($match[1]) {
    case ':d':
      $args[$cnt]= (int)$args[$cnt]; // We must use type casting to int to convert FALSE/NULL/(TRUE?)
      
      if (!$stmt->bindParam(":d". $cnt, $args[$cnt], PDO::PARAM_INT, 32))
      {
  	     ${DB_ERROR} = TRUE;
         trigger_error(check_plain("query: ". $query."bind: :d". $cnt ." ->  ". $args[$cnt]), E_USER_WARNING);
      }
      
      return;
    case ':s':
      $args[$cnt]= oracle_cleanup_arg_value((string)$args[$cnt]); // We must use type casting to string or 0 will match '' empty char

      
      if (!$stmt->bindParam(":s". $cnt, $args[$cnt], PDO::PARAM_STR, strlen($args[$cnt])))
      {
  	     ${DB_ERROR} = TRUE;
         trigger_error(check_plain("query: ". $query."bind: :s". $cnt ." ->  ". $args[$cnt]), E_USER_WARNING);
      }
      
      return;
    case ':n':
      $args[$cnt] = trim($args[$cnt]);
      $args[$cnt] = is_numeric($args[$cnt]) && !preg_match('/x/i', $args[$cnt]) ? $args[$cnt] : '0';

      if (!$stmt->bindParam(":n". $cnt, $args[$cnt], PDO::PARAM_INT, 32))
      {
  	     ${DB_ERROR} = TRUE;
         trigger_error(check_plain("query: ". $query."bind: :n". $cnt ." ->  ". $args[$cnt]), E_USER_WARNING);
      }
      
      return;
    case ':f':
      $args[$cnt]= (string)(float)$args[$cnt];

      if (!$stmt->bindParam(":f". $cnt, $args[$cnt], PDO::PARAM_STR, strlen($args[$cnt])))
      {
  	     ${DB_ERROR} = TRUE;
         trigger_error(check_plain("query: ". $query."bind: :f". $cnt ." ->  ". $args[$cnt]), E_USER_WARNING);
      }
      
      return;
    case ':b': 
      $args[$cnt]= oracle_cleanup_arg_value((string)$args[$cnt]);   
    	
      if (!$stmt->bindParam(":b". $cnt, $args[$cnt], PDO::PARAM_STR, strlen($args[$cnt])))
      {
  	     ${DB_ERROR} = TRUE;
      	 trigger_error(check_plain("query: ". $query."bind: :b". $cnt ." ->  ". $args[$cnt]), E_USER_WARNING);
      }
      
      return;
  }
}


function db_query($query)
{
  $args = func_get_args();
  array_shift($args);
  $query = oracle_format_query($query);
  
  if (isset($args[0]) && is_array($args[0])) 
    $args = array_values($args[0]);
 
  oracle_db_query_callback($args, TRUE);
  $query = preg_replace_callback(DB_QUERY_REGEXP, 'oracle_db_query_callback', $query);
  
  return _db_query($query,$args);
}

function _db_query($query, $args, $debug = 0,$retried=0) {
  global $active_db, $affected_rows, $oracle_error, $oracle_debug, $oracle_unformatted_query;
  
  $commit= true;
  
  try
  {	  
  	  try { $active_db->beginTransaction(); } catch (Exception $ex) {  /* ignore if already in transaction */ $commit= false; }
 
  	  if ($oracle_debug) syslog(LOG_ERR,"query: ".$query." args: ".print_r($args,true));
  	  
	  $stmt = $active_db->prepare($query);
	
	  oracle_bind_query_callback($args,$stmt,TRUE,$query);
	  preg_replace_callback("/(:d|:s|:f|:b|:n)([0-9]+)/", 'oracle_bind_query_callback', $query);
	
	  $stmt->execute();
	  
	  $affected_rows= $stmt->rowCount();
	  $oracle_error= false;
	  
	  if ($commit) 
	    try { $active_db->commit(); } catch (Exception $ex) { /* ignore if not in transaction */ }
	  
	  return $stmt;
  }
  catch (Exception $ex)
  {
  	  if (oracle_exception_query($oracle_unformatted_query)&&$retried!=1)
          {
  	  	 $stmt= _db_query(oracle_retry_query($oracle_unformatted_query,$args), $args, 0, 1);
                 if ($commit)
	             try { $active_db->commit(); } catch (Exception $ex) { /* ignore if not in transaction */ }
                 return $stmt;
          }
  	  
          if (isset($ex->errorInfo)&&is_array($ex->errorInfo)&&$ex->errorInfo[1]=='00972'&&$retried!=2) // catch long identifier errors for alias columns
          {
                 oracle_find_and_remove_long_identifiers($oracle_unformatted_query);
                 $stmt= _db_query(oracle_retry_query($oracle_unformatted_query,$args), $args, 0, 2);
                 if ($commit)
	            try { $active_db->commit(); } catch (Exception $ex) { /* ignore if not in transaction */ }
                 return $stmt;
          }
        	
	  try { $active_db->rollback(); } catch (Exception $ex) { /* ignore if not in transaction */ }
	  
  	  ${DB_ERROR} = TRUE;
	  trigger_error(check_plain($ex->getMessage() ."\nquery: ". $query."\nargs: ".print_r($args,true)).' unformatted: '.$oracle_unformatted_query, E_USER_WARNING);
	  $oracle_error= $ex;
  }
  
  return false;
}

function oracle_retry_query($query,$args)
{
	 $query= oracle_format_query($query);
     oracle_db_query_callback($args, TRUE);
     $query = preg_replace_callback(DB_QUERY_REGEXP, 'oracle_db_query_callback', $query);
     return $query; 	
}

function oracle_exception_query(&$oracle_unformatted_query)
{
	global $oracle_exception_queries;
	
	if (!is_array($oracle_exception_queries)) return false;
	
    $count= 0;	
	$oracle_unformatted_query= preg_replace(array_keys($oracle_exception_queries),array_values($oracle_exception_queries),$oracle_unformatted_query,-1,$count);
	
	return $count;
}

function oracle_find_and_record_long_identifiers($queryPart)
{
  	preg_match_all("/\w+/", $queryPart, $words);
  	$words = $words[0];
     
  	foreach ($words as $word)
  	  if (strlen($word)>ORACLE_IDENTIFIER_MAX_LENGTH)
  	     oracle_identifier($word);
}

function oracle_find_and_remove_long_identifiers($query)
{
  	// do not replace things in literals
  	$literals= array(); 
        $replaceable = preg_split("/'.*?'/", $query);
        
        $lidx= 0;

        // assume that a query cannot start with a literal and that 
        foreach ($replaceable as $toescape)
             oracle_find_and_record_long_identifiers($toescape);

        oracle_reset_long_identifiers(true);
	
}

function oracle_is_long_identifier($key)
{
	return (strtoupper(substr($key,0,ORACLE_LONG_IDENTIFIER_PREFIX_LENGTH))==ORACLE_LONG_IDENTIFIER_PREFIX);
}

function oracle_is_blob($value)
{
	return (substr($value,0,ORACLE_BLOB_PREFIX_LENGTH)==ORACLE_BLOB_PREFIX);
}

function oracle_long_identifier_key($key)
{
	global $hashLongIdentifiers;
	
    return $hashLongIdentifiers[strtoupper($key)];
}

function oracle_cleanup_fetched_value($value)
{
	  if (is_string($value))
  	  {
  	    if ($value==ORACLE_EMPTY_STRING_REPLACER)
  	      return '';
  	    elseif (oracle_is_blob($value))
  	      return oracle_read_blob($value);
  	    else
  	      return $value;
  	  }
  	  else
  	    return $value;
}

function oracle_cleanup_fetched($f)
{
  	    if (is_array($f))
  	      foreach ($f as $key => $value)
  	      {
  	      	if ((string)$key==ORACLE_ROWNUM_ALIAS_LOWER)
  	      		unset($f[$key]);
  	      	elseif (oracle_is_long_identifier($key)) // long identifier
  	      	{
  	      		$f[oracle_long_identifier_key($key)]= oracle_cleanup_fetched($value);
  	      		unset($f[$key]);
  	      	}
  	      	else
  	          $f[$key]= oracle_cleanup_fetched($value);
  	      }
  	    elseif (is_object($f))
          foreach ($f as $key => $value)
  	      {
  	      	if ((string)$key==ORACLE_ROWNUM_ALIAS_LOWER)
  	      		unset($f->{$key});
  	      	elseif (oracle_is_long_identifier($key)) // long identifier
  	      	{
  	      	  $f->{oracle_long_identifier_key($key)}= oracle_cleanup_fetched($value);
  	      	  unset($f->{$key});
  	      	}
  	      	else
  	      	  $f->{$key}= oracle_cleanup_fetched($value);
  	      }
  	    else
  	        $f= oracle_cleanup_fetched_value($f);
  	        
  	    return $f;
}


function db_fetch_object($result)
{
    if ($result)
      return oracle_cleanup_fetched($result->fetchObject());
}

function db_fetch_array($result)
{
    if ($result)
      return oracle_cleanup_fetched($result->fetch(PDO::FETCH_ASSOC));
}

function db_result($result)
{
    if ($result)
      return oracle_cleanup_fetched($result->fetchColumn());
}

/**
 * Determine whether the previous query caused an error.
 */
function db_error() {
  global $oracle_error;
  return $oracle_error;
}

/**
 * Returns the last insert id. This function is thread safe.
 *
 * @param $table
 *   The name of the table you inserted into.
 * @param $field
 *   The name of the autoincrement field.
 */
function db_last_insert_id($table, $field)
{
    $info= oracle_get_table_info($table);
    return oracle_query("select ".$info->sequence_name.".currval from dual")->fetchColumn();    
}

/**
 * Determine the number of rows changed by the preceding query.
 */
function db_affected_rows() {
  global $affected_rows;
  return empty($affected_rows) ? 0 : $affected_rows;
}

function oracle_get_table_schema($table)
{
  	$exp= explode(".",$table);
  	
  	if (count($exp)>1)
  	  return strtoupper(str_replace('"','',$exp[0]));
  	else 
  	  return false;
}

function oracle_reset_long_identifiers ($force=FALSE)
{
	global $foundLongIdentifier,$searchLongIdentifiers,$replaceLongIdentifiers,$hashLongIdentifiers;

    if ($foundLongIdentifier||$force)
    {
      
	    try
	    {
			    $result= oracle_query("select id, identifier from long_identifiers where substr(identifier,1,3) not in ('IDX','TRG','PK_','UK_') order by length(identifier) desc");
			    
			    $searchLongIdentifiers= array();
			    $replaceLongIdentifiers= array();
			    $hashLongIdentifiers= array();
			    
			    while ($row= $result->fetchObject())
			    {
			    	  $searchLongIdentifiers[]= "/".$row->identifier."/i";
			    	  $replaceLongIdentifiers[]= ORACLE_LONG_IDENTIFIER_PREFIX.$row->id;
			    	  $hashLongIdentifiers[ORACLE_LONG_IDENTIFIER_PREFIX.$row->id]= strtolower($row->identifier);
			    }
			    
	    }
	    catch (Exception $ex)
	    {
	    	 // ignore until long_identifiers table is not created
	    }
    	
      $foundLongIdentifier= false;
    }
}

 /**
   * emulates mysql default column behaviour (eg. 
   * insert into table (col1) values (null) 
   * if col1 has default in mysql you have the default insterted instead of null.
   * On oracle you have null inserted.
   * So we need a trigger to intercept this condition and substitute null with default...
   * This condition happens on MySQL only inserting not updating  
   * )
   */
function oracle_rebuild_defaults_trigger_sql($table)
{
  	 $schema= oracle_get_table_schema(oracle_prefix_tables('{' . strtoupper($table) . '}'));
  	 
  	 $oname= oracle_identifier($table,false,false);
  	 
  	 $trigger= 'create or replace trigger '.oracle_identifier('TRG_'.$table.'_DEFS',true).'
                 before insert on '.oracle_identifier($table,true).'
                 for each row
                 begin   /* defs trigger */  
                    
                    if inserting then       
                    ';
  	 
  	 $stmt=   oracle_query("select column_name,
					                                 data_default
					                            from all_tab_columns
					                           where column_name not in (select field_name from table(identifier.get_serial(?,?)))
					                             and owner= nvl(user,?)
					                             and table_name= ?
					                             and data_default is not null",array($table,$schema,$schema,$oname));
  	 
  	 $def= false;
  	 
  	 while ($row=$stmt->fetchObject())
  	 {
  	   $def= true;
  	   $trigger.= '
                                 if :NEW."'.$row->column_name.'" is null or to_char(:NEW."'.$row->column_name.'") = \''.ORACLE_EMPTY_STRING_REPLACER.'\' then
                                    :NEW."'.$row->column_name.'":= '.$row->data_default.';
                                 end if;    
                             ';
  	 }
  	 
  	 if (!$def) $trigger.= '  null; ';
  	  
  	 $trigger.= '
                    end if;
                    
                 end;';
  	 
     return $trigger;
}

function oracle_get_table_info($table)
{
	  global $tableInformation;
	  
      $schema= oracle_get_table_schema(oracle_prefix_tables('{' . strtoupper($table) . '}'));
      
  	  if (!isset($tableInformation[$schema."|".$table]))
  	  {  	  	
  	  	  ////syslog(LOG_ERR,"non cached: ".$table);
  	      $info= new stdClass();
  	      $info->{"sequence_name"} = oracle_query("select identifier.sequence_for_table(?,?) sequence_name from dual",array(strtoupper($table),$schema))->fetchColumn();
  	      
  	      $tableInformation[$schema."|".$table]= $info;
  	  }
  	  else
  	    $info= $tableInformation[$schema."|".$table];
  	  
  	  ////syslog(LOG_ERR,"info: ".print_r($info,true));
  	    
      return $info;
}

function oracle_remove_tableinfo_cache($table)
{
	  global $tableInformation;
      $schema= oracle_get_table_schema(oracle_prefix_tables('{' . strtoupper($table) . '}'));
      unset($tableInformation[$schema."|".$table]);
}
  
/**
 * Runs a limited-range query in the active database.
 *
 * Use this as a substitute for db_query() when a subset of the query
 * is to be returned.
 * User-supplied arguments to the query should be passed in as separate
 * parameters so that they can be properly escaped to avoid SQL injection
 * attacks.
 *
 * @param $query
 *   A string containing an SQL query.
 * @param ...
 *   A variable number of arguments which are substituted into the query
 *   using printf() syntax. Instead of a variable number of query arguments,
 *   you may also pass a single array containing the query arguments.
 *   Valid %-modifiers are: %s, %d, %f, %b (binary data, do not enclose
 *   in '') and %%.
 *
 *   NOTE: using this syntax will cast NULL and FALSE values to decimal 0,
 *   and TRUE values to decimal 1.
 *
 * @param $from
 *   The first result row to return.
 * @param $count
 *   The maximum number of result rows to return.
 * @return
 *   A database query result resource, or FALSE if the query was not executed
 *   correctly.
 */
function db_query_range($query)
{
  $args = func_get_args();
  $count = array_pop($args);
  $from = array_pop($args);
  array_shift($args);

  $query = oracle_format_query($query);
  if (isset($args[0]) and is_array($args[0])) { // 'All arguments in one array' syntax
    $args = array_values($args[0]);
  }
  $query = 'SELECT * FROM (SELECT TAB.*, ROWNUM RWN FROM ('. $query .') TAB) WHERE RWN BETWEEN %d AND %d';
  oracle_db_query_callback($args, TRUE);
  $query = preg_replace_callback(DB_QUERY_REGEXP, 'oracle_db_query_callback', $query);
  
  /* NOTE: "reverse" bind last two parameter so that cases where there are more args than needed still work (eg. menu.inc line 2212:
   * $parent_has_children = (bool)db_result(db_query_range("SELECT mlid FROM {menu_links} WHERE menu_name = '%s' AND plid = %d AND hidden = 0". $where, $item['menu_name'], $item['plid'], $item['mlid'], 0, 1));
   * here the ... AND hidden = 0". $where) hides a parameter when $where is empty but the list of params is fixed...
   */
  $bind_num= oracle_db_query_callback($args, FALSE, TRUE);
  $args[$bind_num-1]= ((int)$from+1);
  $args[$bind_num]= ((int)$count+(int)$from);
  return _db_query($query,$args);
}



/**
 * Runs a SELECT query and stores its results in a temporary table.
 *
 * Use this as a substitute for db_query() when the results need to stored
 * in a temporary table. Temporary tables exist for the duration of the page
 * request.
 * User-supplied arguments to the query should be passed in as separate parameters
 * so that they can be properly escaped to avoid SQL injection attacks.
 *
 * Note that if you need to know how many results were returned, you should do
 * a SELECT COUNT(*) on the temporary table afterwards. db_affected_rows() does
 * not give consistent result across different database types in this case.
 *
 * @param $query
 *   A string containing a normal SELECT SQL query.
 * @param ...
 *   A variable number of arguments which are substituted into the query
 *   using printf() syntax. The query arguments can be enclosed in one
 *   array instead.
 *   Valid %-modifiers are: %s, %d, %f, %b (binary data, do not enclose
 *   in '') and %%.
 *
 *   NOTE: using this syntax will cast NULL and FALSE values to decimal 0,
 *   and TRUE values to decimal 1.
 *
 * @param $table
 *   The name of the temporary table to select into. This name will not be
 *   prefixed as there is no risk of collision.
 * @return
 *   A database query result resource, or FALSE if the query was not executed
 *   correctly.
 */
function db_query_temporary($query)
{
  $args = func_get_args();
  $tablename = array_pop($args);
  $tablename.= "_".oracle_query("SELECT sid from v\$session where audsid = userenv('sessionid')")->fetchColumn();
  array_shift($args);
    
  try { update_sql("DROP TABLE {". $tablename ."}"); } catch (Exception $ex) { /* ignore drop errors */ }
  $result= db_query('CREATE GLOBAL TEMPORARY TABLE {'. $tablename .'} ON COMMIT PRESERVE ROWS AS '.$query, $args);
  
  return array('success' => $result !== FALSE, 'query' => check_plain($sql));
}


/**
 * Returns a properly formatted Binary Large OBject value.
 * In case of Oracle encodes data for insert into BLOB field.
 *
 * @param $data
 *   Data to encode.
 * @return
 *  Encoded data.
 */
function db_encode_blob($data)
{
	return oracle_write_blob($data);
}

/**
 * Returns text from a Binary Large OBject value.
 * In case of Oracle decodes data after select from bytea field.
 *
 * @param $data
 *   Data to decode.
 * @return
 *  Decoded data.
 */
function db_decode_blob($data)
{
    if (oracle_is_blob($data))
      return oracle_read_blob($data);
    else
      return $data;
}

/**
 * Lock a table.
 * This function automatically starts a transaction.
 */
function db_lock_table($table)
{
   global $active_db;
   syslog(LOG_ERR,"start trans");
   try { $active_db->beginTransaction(); } catch (Exception $ex) {  /* ignore if already in transaction */ }
   db_query('LOCK TABLE {'. db_escape_table($table) .'} IN EXCLUSIVE MODE');
}

/**
 * Unlock all locked tables.
 * This function automatically commits a transaction.
 */
function db_unlock_tables()
{
   global $active_db;
   $active_db->commit();
   syslog(LOG_ERR,"stop trans");
}

function oracle_identifier($name, $prefix= FALSE, $quote= TRUE)
{
	  global $foundLongIdentifier;
	  
  	  $ret= $name;
  	  
  	  if (strlen($ret)>ORACLE_IDENTIFIER_MAX_LENGTH)
  	  {
  	  	$foundLongIdentifier= true;
        $ret= oracle_query("select identifier.get_for(?) from dual",array(strtoupper($ret)))->fetchColumn();
  	  }
  	  
  	  if ($prefix)
  	   $ret= "{".$ret."}";
  	  else
  	   $ret= strtoupper($ret);
  	   
  	  if (!$prefix&&$quote)
  	   $ret= '"'.$ret.'"';
  	  
  	  return $ret;
}

/**
 * Check if a table exists.
 */
function db_table_exists($table)
{
  	$schema= oracle_get_table_schema(oracle_prefix_tables('{' . strtoupper($table) . '}'));
  	
  	if ($schema)
  	  return oracle_query("SELECT 1 FROM all_tables WHERE temporary= 'N' and table_name = ? and owner= ?",array(oracle_identifier($table,false,false),$schema))->fetchColumn();
  	else
  	  return oracle_query("SELECT 1 FROM user_tables WHERE temporary= 'N' and table_name = ?",array(oracle_identifier($table,false,false)))->fetchColumn(); 
}

/**
 * Check if a column exists in the given table.
 */
function db_column_exists($table, $column)
{
  	$schema= oracle_get_table_schema(oracle_prefix_tables('{' . strtoupper($table) . '}'));
	
  	if ($schema)
  	  return oracle_query("SELECT 1 FROM all_tab_columns WHERE column_name = ? and table_name = ? and owner= ?",array(oracle_identifier($column,false,false),oracle_identifier($table,false,false),$schema))->fetchColumn();
  	else
  	  return oracle_query("SELECT 1 FROM user_tab_columns WHERE column_name= ? and table_name = ?",array(oracle_identifier($column,false,false),oracle_identifier($table,false,false)))->fetchColumn(); 
  	
}

/**
 * Verify if the database is set up correctly.
 */
function db_check_setup()
{
  $t = get_t();

  $encoding = db_result(db_query("select replace(value,'AL32UTF8','UTF8') from v\$nls_parameters where parameter= 'NLS_CHARACTERSET'"));
  if (!in_array(strtolower($encoding), array('unicode', 'utf8'))) {
    drupal_set_message($t('Your Oracle database is set up with the wrong character encoding (%encoding). It is possible it will not work as expected. It is advised to recreate it with UTF-8/Unicode encoding. More information can be found in the <a href="@url">Oracle documentation</a>.', array('%encoding' => $encoding, '@url' => 'http://download.oracle.com/docs/cd/B19306_01/server.102/b14225/ch6unicode.htm#sthref686')), 'status');
  }
}

/**
 * Moved to database.inc 6.16
 */
if (!function_exists('db_distinct_field'))
{
	function db_distinct_field($table, $field, $query) {
	  if (!preg_match('/FROM\s+\S+\s+AS/si', $query)
	  && !preg_match('/DISTINCT\s+ON\s*\(\s*(' . $table . '\s*\.\s*)?' . $field . '\s*\)/si', $query)
	  && !preg_match('/DISTINCT[ (]' . $field . '/si', $query)
	  && preg_match('/(.*FROM\s+)(.*?\s)(\s*(WHERE|GROUP|HAVING|ORDER|LIMIT|FOR).*)/Asi', $query, $m)) {
	    $query = $m[1];
	    $query .= preg_replace('/([\{\w+\}]+)\s+(' . $table . ')\s/Usi', '(SELECT DISTINCT ON (' . $field . ') * FROM \1) \2 ', $m[2]);
	    $query .= $m[3];
	  }
	  return $query;
	}
}

/**
 * @} End of "ingroup database".
 */

/**
 * @ingroup schemaapi
 * @{
 */

/**
 * This maps a generic data type in combination with its data size
 * to the engine-specific data type.
 */
function db_type_map() {
  // Put :normal last so it gets preserved by array_flip.  This makes
  // it much easier for modules (such as schema.module) to map
  // database types back into schema types.
  $map = array(
    'varchar:normal' => 'varchar2',
    'char:normal' => 'varchar2',

    'text:tiny' => 'varchar2',
    'text:small' => 'varchar2',
    'text:medium' => 'varchar2',
    'text:big' => 'varchar2',
    'text:normal' => 'varchar2',

    'int:tiny' => 'number',
    'int:small' => 'number',
    'int:medium' => 'number',
    'int:big' => 'number',
    'int:normal' => 'number',

    'float:tiny' => 'number',
    'float:small' => 'number',
    'float:medium' => 'number',
    'float:big' => 'number',
    'float:normal' => 'number',

    'numeric:normal' => 'number',

    'blob:big' => 'varchar2',
    'blob:normal' => 'varchar2',

    'datetime:normal' => 'date',

    'serial:tiny' => 'number',
    'serial:small' => 'number',
    'serial:medium' => 'number',
    'serial:big' => 'number',
    'serial:normal' => 'number',
  );
  return $map;
}

function oracle_create_cols_sql($cols)
{
    $ret = array();
    foreach ($cols as $col) $ret[] = oracle_identifier($col);
    return implode(', ', $ret);
}

function oracle_prepare_comment($comment)
{
	$comment = decode_entities(strip_tags($comment));
    return oracle_quote($comment);
}

function oracle_create_serial_sql($table, $field_name, $start_with=1)
{
    $oname= oracle_identifier($table,true);
          
    $trgname= oracle_identifier('TRG_'.$table.'_'.$field_name,true);
    $seqname= oracle_identifier('SEQ_'.$table.'_'.$field_name,true);
    
    $ofield_name= oracle_identifier($field_name);
    
    $tblname_ser= oracle_identifier($table,false,false);
    $trgname_ser= oracle_identifier('TRG_'.$table.'_'.$field_name,false,false);
    $seqname_ser= oracle_identifier('SEQ_'.$table.'_'.$field_name,false,false);
    $fldname_ser= oracle_identifier($field_name,false,false);
    
    $statements[] = 'CREATE SEQUENCE '.$seqname.($start_with!=1 ? ' START WITH '.$start_with : '');
    $statements[] = 'CREATE OR REPLACE TRIGGER '.$trgname.' before insert on '. $oname .'    for each row begin   /* serial('.$tblname_ser.','.$trgname_ser.','.$seqname_ser.','.$fldname_ser.') */  if inserting then       if :NEW.'. $ofield_name .' is null or :NEW.'. $ofield_name .' = 0 then   select '. $seqname .'.nextval into :NEW.'. $ofield_name .' from dual;       end if;    end if; end;';
    $statements[] = 'ALTER TRIGGER '.$trgname.' ENABLE';
        
    return $statements; 
}

/**
 * Generate SQL to create a new table from a Drupal schema definition.
 *
 * @param $name
 *   The name of the table to create.
 * @param $tableselect distinct data_type from user_tab_columns
 *   A Schema API table definition array.
 * @return
 *   An array of SQL statements to create the table.
 */
function db_create_table_sql($name, $table)
{
  	$oname= oracle_identifier($name,true);
  	
  	////syslog(LOG_ERR,"oname: ".$oname);
  	
    $sql_fields = array();
    foreach ($table['fields'] as $field_name => $field) {
      $sql_fields[] = _db_create_field_sql($field_name, _db_process_field($field));
    }

    $sql_keys = array();
    
    if (isset($table['primary key']) && is_array($table['primary key']))
       $sql_keys[]= 'CONSTRAINT '.oracle_identifier('PK_'.$name).' PRIMARY KEY ('.oracle_create_cols_sql($table['primary key']).')';
   
    
    if (isset($table['unique keys']) && is_array($table['unique keys']))
      foreach ($table['unique keys'] as $key_name => $key)
        $sql_keys[]= 'CONSTRAINT '.oracle_identifier('UK_'.$name.'_'.$key_name).' UNIQUE ('.oracle_create_cols_sql($key).')';
      	
    
    $sql = "CREATE TABLE " . $oname . " (\n\t";
    $sql .= implode(",\n\t", $sql_fields);
    if (count($sql_keys) > 0) {
      $sql .= ",\n\t";
    }
    $sql .= implode(",\n\t", $sql_keys);
    $sql .= "\n)";
    $statements[] = $sql;

    if (isset($table['indexes']) && is_array($table['indexes'])) 
      foreach ($table['indexes'] as $key_name => $key) 
        $statements[] = _db_create_index_sql($name, $key_name, $key);
    
    // Add table comment.
    if (isset($table['description'])&&oracle_prepare_comment($table['description'])) 
      $statements[] = 'COMMENT ON TABLE ' . $oname . ' IS ' . oracle_prepare_comment($table['description']);
    

    // Add column comments.
    foreach ($table['fields'] as $field_name => $field) 
      if (isset($field['description'])&&oracle_prepare_comment($field['description'])) 
        $statements[] = 'COMMENT ON COLUMN ' . $oname . '.' . oracle_identifier($field_name) . ' IS ' . oracle_prepare_comment($field['description']);
      
	foreach ($table['fields'] as $field_name => $field)
	{
	
	    if ($field['type'] == 'serial')
	    	$statements= array_merge($statements,	oracle_create_serial_sql($name,$field_name));
	    elseif ($field['type'] == 'blob')
	        $statements[] = "INSERT INTO BLOB_COLUMN VALUES ('". strtoupper($name) ."','". strtoupper($field_name) ."')";
    }     

    $statements[] = oracle_rebuild_defaults_trigger_sql($name);
    
    oracle_reset_long_identifiers();
    
    return $statements;
}

function _db_create_index_sql($table, $name, $fields)
{
  	$oname= oracle_identifier('IDX_'.$table.'_'.$name,true);
  	
    $query = 'CREATE INDEX '.$oname.' ON ' . oracle_identifier($table,true) . ' (';
    $query .= _db_create_key_sql($fields) . ')';
    
    return $query;
}

function _db_create_key_sql($fields)
{
    $ret = array();
    
    foreach ($fields as $field) 
      if (is_array($field)) 
        $ret[] = 'substr(' . oracle_identifier($field[0]) . ', 1, ' . $field[1] . ')';
      else 
        $ret[] = oracle_identifier($field);
      
    
    return implode(', ', $ret);
}

function _db_create_keys(&$ret, $table, $new_keys)
{
  if (isset($new_keys['primary key'])) {
    db_add_primary_key($ret, $table, $new_keys['primary key']);
  }
  if (isset($new_keys['unique keys'])) {
    foreach ($new_keys['unique keys'] as $name => $fields) {
      db_add_unique_key($ret, $table, $name, $fields);
    }
  }
  if (isset($new_keys['indexes'])) {
    foreach ($new_keys['indexes'] as $name => $fields) {
      db_add_index($ret, $table, $name, $fields);
    }
  }
}

/**
 * Set database-engine specific properties for a field.
 *
 * @param $field
 *   A field description array, as specified in the schema documentation.
 */
function _db_process_field($field)
{
  	if (!isset($field['size']))
      $field['size'] = 'normal';
    
    // Set the correct database-engine specific datatype.
    if (!isset($field['oracle_type'])) {
      $map = db_type_map();
      $field['oracle_type'] = $map[$field['type'] . ':' . $field['size']];
    }
    
    return $field;
}

function oracle_quote($string)
{
	return "'".db_escape_string($string)."'";
}

/**
 * Create an SQL string for a field to be used in table creation or alteration.
 *
 * Before passing a field out of a schema definition into this function it has
 * to be processed by _db_process_field().
 *
 * @param $name
 *    Name of the field.
 * @param $spec
 *    The field specification, as per the schema data structure format.
 */
function _db_create_field_sql($name, $spec)
{
  	$oname= oracle_identifier($name);
    
  	$sql = $oname . ' ' . $spec['oracle_type'];

    if ($spec['type'] == 'serial') {
      unset($spec['not null']);
    }
    
    
    if ($spec['oracle_type']=='varchar2')
    {
      $sql .= '('.(!empty($spec['length']) ? $spec['length'] : ORACLE_MAX_VARCHAR2_LENGTH ).' CHAR)';
    }
    elseif (!empty($spec['length'])) {
      $sql .= '(' . $spec['length'] . ')';
    }
    elseif (isset($spec['precision']) && isset($spec['scale'])) {
      $sql .= '(' . $spec['precision'] . ', ' . $spec['scale'] . ')';
    }

    if (isset($spec['default'])) {
      $default = is_string($spec['default']) ? oracle_quote(oracle_cleanup_arg_value($spec['default'])) : $spec['default'];
      
      $sql .= " default $default";
    }

    if (isset($spec['not null']) && $spec['not null']) {
      $sql .= ' NOT NULL';
    }
    
    return $sql;
}

function oracle_table_serial_info($table)
{
    $schema= oracle_get_table_schema(oracle_prefix_tables('{' . strtoupper($table) . '}'));
  	return oracle_query("select * from table(identifier.get_serial(?, ?))",array(strtoupper($table),$schema))->fetchObject();
}

/**
 * Rename a table.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   The table to be renamed.
 * @param $new_name
 *   The new name for the table.
 */
function db_rename_table(&$ret, $table, $new_name)
{
    $info= oracle_table_serial_info($table);
  	$oname= oracle_identifier($new_name,true);
  	
  	if (isset($info->sequence_name)&&$info->sequence_name)
  	{
  	   $ret[] = update_sql('DROP TRIGGER {'.$info->trigger_name.'}');
  	   $ret[] = update_sql('DROP SEQUENCE {'.$info->sequence_name.'}');
  	}
  	
  	$ret[] = update_sql("DROP TRIGGER ".oracle_identifier('TRG_'.$table.'_DEFS',true)); // drop defaults trigger
  	
    $ret[] = update_sql('ALTER TABLE ' . oracle_identifier($table,true) . ' RENAME TO ' . oracle_identifier($new_name,false) ); // should not use prefix because schema is not needed on rename
    
  	if (isset($info->sequence_name)&&$info->sequence_name)
  	{
  	    $statements= oracle_create_serial_sql($table,$info->field_name,$info->sequence_restart);
  	    
  	    foreach ($statements as $statement)
  	     $ret[] = update_sql($statement);
  	}
    
  	$ret[] = update_sql(oracle_rebuild_defaults_trigger_sql($new_name));
  	
  	oracle_reset_long_identifiers();
    oracle_remove_tableinfo_cache($table);
	
}

/**
 * Drop a table.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   The table to be dropped.
 */
function db_drop_table(&$ret, $table)
{
  	$info= oracle_get_table_info($table);
  	
  	if ($info->sequence_name)
  	  $ret[] = update_sql('DROP SEQUENCE '.$info->sequence_name);
  	
    $ret[] = update_sql('DROP TABLE ' . oracle_identifier($table,true) . ' CASCADE CONSTRAINTS');
    
    oracle_remove_tableinfo_cache($table);
}

/**
 * Add a new field to a table.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   Name of the table to be altered.
 * @param $field
 *   Name of the field to be added.
 * @param $spec
 *   The field specification array, as taken from a schema definition.
 *   The specification may also contain the key 'initial', the newly
 *   created field will be set to the value of the key in all rows.
 *   This is most useful for creating NOT NULL columns with no default
 *   value in existing tables.
 * @param $keys_new
 *   Optional keys and indexes specification to be created on the
 *   table along with adding the field. The format is the same as a
 *   table specification but without the 'fields' element.  If you are
 *   adding a type 'serial' field, you MUST specify at least one key
 *   or index including it in this array. @see db_change_field for more
 *   explanation why.
 */
function db_add_field(&$ret, $table, $field, $spec, $new_keys = array())
{
    $fixnull = FALSE;
    
    if (!empty($spec['not null']) && !isset($spec['default'])) {
      $fixnull = TRUE;
      $spec['not null'] = FALSE;
    }
    
    $query = 'ALTER TABLE ' . oracle_identifier($table,true) . ' ADD (';
    
    $query .= _db_create_field_sql($field, _db_process_field($spec));
    
    $query .= ')';
    
    $ret[] = update_sql($query);
    
    if (isset($spec['initial'])) {
      // All this because update_sql does not support %-placeholders.
      $sql = 'UPDATE ' . oracle_identifier($table,true) . ' SET ' . oracle_identifier($field) . ' = ' . db_type_placeholder($spec['type']);
      $result = db_query($sql, $spec['initial'] );
      
      $ret[] = array('success' => $result !== FALSE, 'query' => check_plain($sql . ' (' . $spec['initial'] . ')'));
      
    }
    
    if ($fixnull) 
      $ret[] = update_sql("ALTER TABLE " . oracle_identifier($table,true) . " MODIFY (". oracle_identifier($field) ." NOT NULL)");
    
    
    if (isset($new_keys)) 
      _db_create_keys($ret, $table, $new_keys);
    

    // Add column comment.
    if (!empty($spec['description'])) 
      $ret[] = update_sql('COMMENT ON COLUMN ' . oracle_identifier($table,true) . '.' . oracle_identifier($field) . ' IS ' . oracle_prepare_comment($spec['description']));
    
    
    if ($spec['type'] == 'serial')
    {
	    $statements= oracle_create_serial_sql($table,$field);
	    	
  	    foreach ($statements as $statement)
  	     $ret[] = update_sql($statement);	
    }
    
    $ret[] = update_sql(oracle_rebuild_defaults_trigger_sql($table));
    
	oracle_remove_tableinfo_cache($table);    
    oracle_reset_long_identifiers();
    
}

/**
 * Drop a field.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   The table to be altered.
 * @param $field
 *   The field to be dropped.
 */
function db_drop_field(&$ret, $table, $field)
{
  	$info= oracle_table_serial_info($table);
  	
  	if (isset($info->sequence_name)&&$info->sequence_name&&oracle_identifier($field,false,false)==$info->field_name)
  	{
  	   $ret[] = update_sql('DROP TRIGGER {'.$info->trigger_name.'}');
  	   $ret[] = update_sql('DROP SEQUENCE {'.$info->sequence_name.'}');
  	}
  	
    $ret[] = update_sql('ALTER TABLE ' . oracle_identifier($table,true) . ' DROP COLUMN ' . oracle_identifier($field));
    
    $ret[] = update_sql(oracle_rebuild_defaults_trigger_sql($table));
    
	oracle_remove_tableinfo_cache($table);    
    oracle_reset_long_identifiers();
}

/**
 * Set the default value for a field.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   The table to be altered.
 * @param $field
 *   The field to be altered.
 * @param $default
 *   Default value to be set. NULL for 'default NULL'.
 */
function db_field_set_default(&$ret, $table, $field, $default)
{
    if (is_null($default)) 
      $default = 'NULL';
    else
      $default = is_string($default) ? oracle_quote(oracle_cleanup_arg_value($default)) : $default;

    $ret[] = update_sql('ALTER TABLE ' . oracle_identifier($table,true) . ' MODIFY (' . oracle_identifier($field) . ' DEFAULT ' . $default .' )');
    $ret[] = update_sql(oracle_rebuild_defaults_trigger_sql($table));
}

/**
 * Set a field to have no default value.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   The table to be altered.
 * @param $field
 *   The field to be altered.
 */
function db_field_set_no_default(&$ret, $table, $field)
{
    $ret[] = update_sql('ALTER TABLE ' . oracle_identifier($table,true) . ' MODIFY (' . oracle_identifier($field) . ' DEFAULT NULL)');
    $ret[] = update_sql(oracle_rebuild_defaults_trigger_sql($table));
}

/**
 * Add a primary key.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   The table to be altered.
 * @param $fields
 *   Fields for the primary key.
 */
function db_add_primary_key(&$ret, $table, $fields)
{
   $ret[] = update_sql('ALTER TABLE ' . oracle_identifier($table,true) . ' ADD CONSTRAINT '.oracle_identifier('PK_'.$table).' PRIMARY KEY (' . oracle_create_cols_sql($fields) . ')');
}

/**
 * Drop the primary key.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   The table to be altered.
 */
function db_drop_primary_key(&$ret, $table)
{
    $ret[] = update_sql('ALTER TABLE ' . oracle_identifier($table,true) . ' DROP CONSTRAINT '.oracle_identifier('PK_'.$table));
}

/**
 * Add a unique key.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   The table to be altered.
 * @param $name
 *   The name of the key.
 * @param $fields
 *   An array of field names.
 */
function db_add_unique_key(&$ret, $table, $name, $fields)
{
	$ret[] = update_sql('ALTER TABLE ' . oracle_identifier($table,true) . ' ADD CONSTRAINT ' . oracle_identifier('UK_'.$table.'_'.$name) . ' UNIQUE (' . oracle_create_cols_sql($fields) . ')');
}

/**
 * Drop a unique key.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   The table to be altered.
 * @param $name
 *   The name of the key.
 */
function db_drop_unique_key(&$ret, $table, $name)
{
    $ret[] = update_sql('ALTER TABLE ' . oracle_identifier($table,true) . ' DROP CONSTRAINT ' . oracle_identifier('UK_'.$table.'_'.$name));
}

/**
 * Add an index.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   The table to be altered.
 * @param $name
 *   The name of the index.
 * @param $fields
 *   An array of field names.
 */
function db_add_index(&$ret, $table, $name, $fields)
{
   $ret[] = update_sql(_db_create_index_sql($table, $name, $fields));
}

/**
 * Drop an index.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   The table to be altered.
 * @param $name
 *   The name of the index.
 */
function db_drop_index(&$ret, $table, $name)
{
   $ret[] = update_sql('DROP INDEX ' . oracle_identifier('IDX_'.$table.'_'.$name,true));
}

/**
 * Change a field definition.
 *
 * IMPORTANT NOTE: To maintain database portability, you have to explicitly
 * recreate all indices and primary keys that are using the changed field.
 *
 * That means that you have to drop all affected keys and indexes with
 * db_drop_{primary_key,unique_key,index}() before calling db_change_field().
 * To recreate the keys and indices, pass the key definitions as the
 * optional $new_keys argument directly to db_change_field().
 *
 * For example, suppose you have:
 * @code
 * $schema['foo'] = array(
 *   'fields' => array(
 *     'bar' => array('type' => 'int', 'not null' => TRUE)
 *   ),
 *   'primary key' => array('bar')
 * );
 * @endcode
 * and you want to change foo.bar to be type serial, leaving it as the
 * primary key.  The correct sequence is:
 * @code
 * db_drop_primary_key($ret, 'foo');
 * db_change_field($ret, 'foo', 'bar', 'bar',
 *   array('type' => 'serial', 'not null' => TRUE),
 *   array('primary key' => array('bar')));
 * @endcode
 *
 * The reasons for this are due to the different database engines:
 *
 * On Oracle, changing a field definition involves adding a new field
 * and dropping an old one which* causes any indices, primary keys and
 * sequences (from serial-type fields) that use the changed field to be dropped.
 *
 * On MySQL, all type 'serial' fields must be part of at least one key
 * or index as soon as they are created.  You cannot use
 * db_add_{primary_key,unique_key,index}() for this purpose because
 * the ALTER TABLE command will fail to add the column without a key
 * or index specification.  The solution is to use the optional
 * $new_keys argument to create the key or index at the same time as
 * field.
 *
 * You could use db_add_{primary_key,unique_key,index}() in all cases
 * unless you are converting a field to be type serial. You can use
 * the $new_keys argument in all cases.
 *
 * @param $ret
 *   Array to which query results will be added.
 * @param $table
 *   Name of the table.
 * @param $field
 *   Name of the field to change.
 * @param $field_new
 *   New name for the field (set to the same as $field if you don't want to change the name).
 * @param $spec
 *   The field specification for the new field.
 * @param $new_keys
 *   Optional keys and indexes specification to be created on the
 *   table along with changing the field. The format is the same as a
 *   table specification but without the 'fields' element.
 */
function db_change_field(&$ret, $table, $field, $field_new, $spec, $new_keys = array())
{
  	$info= oracle_table_serial_info($table);
  	
  	if (isset($info->sequence_name)&&$info->sequence_name&&oracle_identifier($field,false,false)==$info->field_name)
  	{
  	   $ret[] = update_sql('DROP TRIGGER {'.$info->trigger_name.'}');
  	   $ret[] = update_sql('DROP SEQUENCE {'.$info->sequence_name.'}');
  	}
  	
  	
    $ret[] = update_sql("ALTER TABLE " . oracle_identifier($table,true) . " RENAME COLUMN ". oracle_identifier($field) ." TO " . oracle_identifier($field.'_old'));
    
    $not_null = isset($spec['not null']) ? $spec['not null'] : FALSE;
    
    unset($spec['not null']);

    if (!array_key_exists('size', $spec)) {
      $spec['size'] = 'normal';
    }
    
    db_add_field($ret, $table, "$field_new", $spec);

    $map = db_type_map();
    $ret[] = update_sql("UPDATE " . oracle_identifier($table,true) . " SET ". oracle_identifier($field_new) . " = " . oracle_identifier($field.'_old'));

    if ($not_null) {
      $ret[] = update_sql("ALTER TABLE " . oracle_identifier($table,true) . " MODIFY (". oracle_identifier($field_new) . " NOT NULL)");
    }

    db_drop_field($ret, $table, $field . '_old');

    if (isset($new_keys)) 
      _db_create_keys($ret, $table, $new_keys);
    
    
  	if (isset($info->sequence_name)&&$info->sequence_name&&oracle_identifier($field,false,false)==$info->field_name)
  	{
  	    $statements= oracle_create_serial_sql($table,oracle_identifier($field_new,false,false),$info->sequence_restart);
  	    
  	    foreach ($statements as $statement)
  	     $ret[] = update_sql($statement);
  	}
    
    $ret[] = update_sql(oracle_rebuild_defaults_trigger_sql($table));
      	
    oracle_remove_tableinfo_cache($table);
    oracle_reset_long_identifiers();
}

/**
 * @} End of "ingroup schemaapi".
 */

closelog();

?>
